"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const model_definitions_1 = require("./model_definitions");
const mongodb_1 = require("mongodb");
class Model {
    constructor(_sm, _definition) {
        this._sm = _sm;
        this._definition = _definition;
        this._new = new WeakSet();
        this._fetched = new WeakMap();
        model_definitions_1.validate(_definition);
        this._version = _definition.versions.length - 1;
        this._collection = _sm.awaitConnected().then(() => _sm.db.collection(_definition.name));
    }
    get version() {
        return this._version;
    }
    new(data) {
        if (data)
            this._validate(data, false, false, true);
        let res = {
            _v: this._version
        };
        this._validate(res, false, true, true);
        if (data) {
            for (let key in data) {
                res[key] = data[key];
            }
        }
        this._new.add(res);
        return res;
    }
    validate(data) {
        this._validate(data, false);
    }
    async save(data) {
        let old = undefined;
        if (!this._new.has(data) && !(old = this._fetched.get(data))) {
            throw new Error("This object was not initialized by using Model.new or prevoiously fetched!");
        }
        this._validate(data, !!old);
        data._v = this._version;
        await this._sm.awaitConnected();
        if (!old) { // Is new
            data._id = undefined; //make shure no invalid IDs are passed
            await (await this._collection).insertOne(data).then(inserted => {
                this._new.delete(data);
                data._id = inserted.insertedId;
                this._fetched.set(data, recursiveDeepCopy(data));
                return data;
            });
        }
        else {
            await (await this._collection).replaceOne({ _id: data._id }, data).then(replace => data);
        }
    }
    async findById(id) {
        if (typeof id === "string")
            id = new mongodb_1.ObjectID(id);
        let res = await (await this._collection).findOne({ _id: id });
        if (!res)
            return null;
        res = await this._upgrade(res);
        this._add_fetched(res);
        return res;
    }
    async findOne(filter) {
        let res = await (await this._collection).findOne(filter);
        if (!res)
            return null;
        res = await this._upgrade(res);
        this._add_fetched(res);
        return res;
    }
    async find(filter) {
        let res = await (await this._collection).find(filter);
        let elms = await res.toArray();
        elms = await Promise.all(elms.map(e => this._upgrade(e)));
        elms.forEach(e => this._add_fetched(e));
        return elms;
    }
    _add_fetched(data) {
        this._fetched.set(data, recursiveDeepCopy(data));
    }
    async _upgrade(data) {
        // console.log("uprade", data._v, this._version);
        if (data._v != this._version) {
            if (data._v > this._version)
                throw new Error("Object version is larger, than this application supports. Please upgrade!");
            let versions = this._definition.versions.slice(data._v, this._definition.versions.length);
            for (let version of versions) {
                if (version.migration) {
                    await version.migration(data); //await don't cares if it is a promise or not
                }
            }
        }
        data._v = this._version;
        return data;
    }
    _validate(data, check_id = true, add_default = false, all_optional = false) {
        if (check_id) {
            if (!mongodb_1.ObjectID.isValid(data._id))
                throw new Error("Invalid _id");
        }
        let version = getLast(this._definition.versions);
        const checkObj = (obj, schema) => {
            let additional = Object.keys(obj).filter(v => !schema[v] && v !== "_id" && v !== "_v");
            if (additional.length > 0)
                throw new Error("Invalid properties set! " + additional.join(" "));
            for (let key in schema) {
                if (key === "_id" || key === "_v")
                    continue;
                let should = schema[key];
                let val = obj[key];
                if (val) {
                    if (should.model) {
                        checkObj(val, should.type);
                    }
                    else {
                        switch (should.type) {
                            case String:
                                if (typeof val !== "string")
                                    throw new Error(key + " should be of type string");
                                break;
                            case Number:
                                if (typeof val !== "number")
                                    throw new Error(key + " should be of type number");
                                break;
                            case Boolean:
                                if (typeof val !== "boolean")
                                    throw new Error(key + " should be of type boolean");
                                break;
                            case Date:
                            case Array:
                                if (!(val instanceof should.type))
                                    throw new Error(key + " should be of type " + should.type.name);
                                break;
                            case mongodb_1.ObjectID:
                                if (!mongodb_1.ObjectID.isValid(val))
                                    throw new Error(key + " should be of type " + should.type.name);
                                break;
                            default:
                                throw new Error(key + " invalid datatype!");
                        }
                    }
                }
                else {
                    if (should.model) {
                        if (!obj[key])
                            obj[key] = {};
                        checkObj(obj[key], should.type);
                    }
                    else if (add_default && should.default) {
                        let def;
                        // console.log(typeof should.default === "function" ? should.default.apply(obj) : "nofunction");
                        if (typeof should.default === "function")
                            def = should.default.apply(obj);
                        else
                            def = should.default;
                        obj[key] = def;
                    }
                    else if (!should.optional && !all_optional)
                        throw new Error(key + " should be defined");
                }
            }
        };
        checkObj(data, version.schema);
    }
}
exports.default = Model;
function getLast(arr) {
    return arr[arr.length - 1];
}
function recursiveDeepCopy(o) {
    var newO, i;
    if (typeof o !== 'object') {
        return o;
    }
    if (!o) {
        return o;
    }
    if (Array.isArray(o)) {
        newO = [];
        for (i = 0; i < o.length; i += 1) {
            newO[i] = recursiveDeepCopy(o[i]);
        }
        return newO;
    }
    newO = {};
    for (i in o) {
        if (o.hasOwnProperty(i)) {
            newO[i] = recursiveDeepCopy(o[i]);
        }
    }
    return newO;
}
exports.recursiveDeepCopy = recursiveDeepCopy;
//# sourceMappingURL=model.js.map